/**
 * Given an (eps,delta)-approximation to the collection of Frequent Itemsets, a
 * dataset, and a minimum frequency threshold, compute the subset of candidates
 * which are indeed frequent in the dataset
 *
 * Copyright 2014 Matteo Riondato <matteo@cs.brown.edu>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * 
 */

#include <algorithm>
#include <cerrno>
#include <chrono>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <iostream>
#include <map>
#include <set>

#include "util.h"

using namespace std;
using namespace std::chrono;

int main(int argc, char **argv) {
	if (argc != 4) {
		cerr << argv[0] << ": filter a set of candidates to find those with frequency at least theta in dataset" << endl; 
		cerr << "USAGE: " << argv[0] << " theta candidates_file dataset" << endl;
		return 1;
	}
	double theta = strtod(argv[1], NULL);
	if (errno == ERANGE || theta < 0.0 || theta > 1.0) {
		cerr << ERROR_HEADER << "theta must be between 0.0 and 1.0"
			<< endl;
		return 1;
	}

	// Open candidates file
	FILE *cand_FILE = fopen(argv[2], "r");
	if (cand_FILE == NULL) {
		perror("Error opening candidates file");
		return errno;
	}
	map<set<unsigned int>,unsigned int> counters;
	char line[TRANSACTION_LINE_MAXLEN];
	// Ignore the first line (size of the sample)
	get_next_line(line, sizeof(line), cand_FILE);
	while (get_next_line(line, sizeof(line), cand_FILE) != false) {
		char *copy = (char *) malloc_check(strlen(line) + 1, "main");
		char *orig_copy = copy;
		strncpy(copy, line, strlen(line) + 1);
		char *itemset_str = strsep(&copy, "(");
		set<unsigned int> itemset = transaction_to_set(itemset_str);
		free(orig_copy);
		counters[itemset] = 0;
	}
	// Close candidates file
	fclose(cand_FILE);

	// We start counting time from here because ideally this procedure would be
	// integrated in the mining process
	auto start = steady_clock::now();

	// Open dataset file
	FILE *ds_FILE = fopen(argv[3], "r");
	if (ds_FILE == NULL) {
		perror("Error opening dataset file");
		return errno;
	}

	unsigned int dataset_size = 0;
	while (get_next_line(line, sizeof(line), ds_FILE) != false) {
		++dataset_size;
		set<unsigned int> transaction = transaction_to_set(line);
		for (auto ic_pair : counters) {
			if (includes(transaction.begin(), transaction.end(),
						ic_pair.first.begin(),
						ic_pair.first.end())) {
				counters[ic_pair.first]++;
			}
		}
	}

	// We stop the clock here because all the rest is not necessarily part
	// of the mining process
	auto end = steady_clock::now();
	auto elapsed = duration_cast<milliseconds>(end - start);

	// Close dataset file
	fclose(ds_FILE);

	unsigned int min_support = (unsigned int) ceil(dataset_size * theta);
	for (auto ic_pair : counters) {
		if (ic_pair.second >= min_support) {
			cout << itemset_to_str(ic_pair.first) << " (" <<
				ic_pair.second << ")" << endl;
		}
	}

	// Printing useful information to stderr
	cerr << "dataset,theta,candidates,filtering_time" << endl;
	cerr << (strrchr(argv[3], '/') + 1) << COMMA << theta << COMMA <<
		(strrchr(argv[2], '/') + 1) << COMMA << elapsed.count();

	return 0;
}

